;;; -*- mode: LISP; package: CLONES -*-

(in-package :clones)


(defgeneric cells-similar-p (a b)
  (:documentation "Returns true if two cells A and B are similar.")
  (:method (a b)
   "No cells are similar by default."
   nil))


(defclass field ()
  ((field-width :initarg :field-width
                :reader field-width
                :initform *field-width*)
   (field-height :initarg :field-height
                 :reader field-height
                 :initform *field-height*)
   (cells)))


;;; Color is no longer an integer; should update to
;;; show character representations for colors.
;;; 
;; (defun show-field (field &optional (stream *standard-output*))
;;   (with-slots (cells) field
;;     (let ((height (length (aref cells 0))))
;;       ;; Show column numbers
;;       (loop for i below (length cells)
;;             initially (format stream " ")
;;             do (format stream " ~2D" i)
;;             finally (terpri stream))
;;       ;; Show contents with row numbers
;;       (loop for j below height
;;             do (format stream "~2D" j)
;;                (loop for column across cells
;;                      for cell = (aref column j)
;;                      do (format stream "~:[ ~C ~;(~C)~]"
;;                                 (cell-highlight-p cell)
;;                                 (schar " *@." (cell-color cell))))
;;                (terpri stream)))))


(defmethod initialize-instance :after ((object field) &rest initargs)
  (declare (ignore initargs))
  (initialize-field object))


(defun initialize-field (field)
  (with-slots (field-width field-height cells) field
    (setf cells (make-cell-grid field field-width field-height))))


(defun make-cell-grid (field width height)
  (let ((columns 
         (loop for i below width
               collect (let ((cells (loop for j below height
                                          collect (make-instance 'cell
                                                    :x i
                                                    :y j
                                                    :color (svref *colors* (random +color-count+))
                                                    :field field))))
                         (make-array height
                                     :element-type 'cell
                                     :initial-contents cells)))))
    (make-array width
                :element-type t
                :initial-contents columns)))


(defun %cell-at (cells x y)
  (aref (aref cells x) y))


(defun cell-at (field x y)
  (with-slots (cells) field
    (%cell-at cells x y)))


(defun field-empty-p (field)
  (every #'column-empty-p (slot-value field 'cells)))


(defmethod cells-similar-p ((a cell) (b cell))
  (eql (cell-color a) (cell-color b)))


(defmethod cells-similar-p ((a integer) (b integer))
  (= a b))


(defun moves-left-p (field)
  "Returns true if there is at least one region left, false otherwise."
  ;; Start from bottom-left corner 
  (let* ((cells (slot-value field 'cells))
         (width (length cells))
         (height (length (aref cells 0))))
    (loop
     for i below (1- width)
     until (cell-empty-p (%cell-at cells i (1- height)))
     do (loop
         for j downfrom (1- height) to 1
         for cell = (%cell-at cells i j)
         do (cond ((cell-empty-p cell)
                   (return))
                  ((or (cells-similar-p cell (%cell-at cells i (1- j)))
                       (cells-similar-p cell (%cell-at cells (1+ i) j)))
                   (return-from moves-left-p t)))))
    nil))



(defun cell-in-list-p (cell list)
  #-(and)
  (find-if #'(lambda (item)
               (same-cell-p item cell))
           list)
  (find cell list :test #'eq))



(defun similar-cells (cells x y)
  "Returns a list of cells that are similar to the cell at X, Y in the
CELLS array.  This always contains at least one cell that has the
given X and Y.  This function is for internal use.  See function
REGION-AT."
  (let ((result ())
        (reference (%cell-at cells x y)))
    (labels ((test (x y)
               ;;(format t "~&X:~D, Y:~D~%" x y)
               (let ((cell (%cell-at cells x y)))
                 (when (and (cells-similar-p cell reference)
                            (not (cell-in-list-p cell result)))
                   (push cell result)
                   ;; now process neighbors
                   (loop
                    for (i . j) in '((0 . -1) (0 . 1) (-1 . 0) (1 . 0))
                    for x1 = (+ x i)
                    for y1 = (+ y j)
                    do (when (and (<= 0 x1 (1- (length cells)))
                                  (<= 0 y1 (1- (length (aref cells x1)))))
                         (test x1 y1)))))))
      (test x y))
    result))


(defun region-at (field x y)
  "Returns a list of cells for a region of similar cells.  Region has
two or more cells (i.e., if there are no similar cells next to the one
by given coordinates, nil is returned)."
  (let ((result (similar-cells (slot-value field 'cells) x y)))
    (if (null (cdr result))
        nil
        result)))


(defun set-region-highlight (list value)
  (loop for cell in list
        do (setf (cell-highlight-p cell) value))
  value)


(defun remove-cells (field list)
  "Makes cells inactive by changing them to empty cells."
  (declare (ignore field))
  (dolist (cell list)
    (setf (cell-color cell) :empty)))


(defun column-empty-p (array)
  "For internal use.  Returns true if all cells in column are empty."
  (null (position-if-not #'cell-empty-p array)))


(defun collapse-array (array test from-end)
  "Collapses ARRAY by moving all cells for which TEST function returns
true after other cells.  Returns true if all cells satisfy TEST
function (this means false for an empty array).  If FROM-END is true,
non-empty cells are moved to the end of the array, if false -- to the
beginning."
  (let ((first (position-if test array :from-end from-end)))
    (cond ((null first)
           ;; no cells satisfy TEST => nothing to do
           nil)

          (t
           (let* ((direction (if from-end -1 1))
                  (i first)
                  (j (+ i direction))
                  (length (length array)))
             (declare (type fixnum direction i j length))
             (loop
              (cond ((or (< j 0)
                         (<= length j))
                     ;; The end reached
                     (return))
                    ((funcall test (aref array j))
                     ;; Cell satisfies test => move over
                     (incf j direction))
                    (t
                     ;; Test not satisfied => swap cells
                     (rotatef (aref array i) (aref array j))
                     (incf i direction)
                     (incf j direction))))
             (zerop i))))))


(defun normalize-field (field)
  "Makes sure the field is in a normalized field.  A normalized field
is one where there are no empty columns between non-empty ones, and
where there are no 'hanging' cells."
  (with-slots (cells) field
    ;; normalize columns
    (loop for column across cells
          do (normalize-column column))
    ;; move all empty columns to the end
    (collapse-array cells #'column-empty-p nil)
    ;; Update cell positions (some cells might have moved)
    (loop for column across cells
          for x upfrom 0
          do (loop for cell across column
                   for y upfrom 0
                   do (setf (cell-x cell) x
                            (cell-y cell) y)))))


(defun normalize-column (array)
  (collapse-array array #'cell-empty-p t))


#|
;;; This is actually quite internal to SHIFT-BLOCK-SWAP
;;;

(defun normalize-column (array)
  "Moves empty cells to the top of the array.  Returns true if all cells are empty."

  (let ((p1 (position 0 array :test-not #'eql)))
    (when (not p1)
      ;; empty column
      (return-from normalize-column t))

    (loop
     (let ((p2 (position 0 array :start p1)))
       (when (not p2)
         ;; nothing to do -- the column is already normalized
         (return-from normalize-column nil))
       
       (let ((p3 (position 0 array :start p2 :test-not #'eql))
             (shift (- p2 p1)))
         ;; p1 - first non-empty cell
         ;; p2 - end of non-empty cells / start of empty cells
         ;; p3 - end of empty cells that
         ;; must shift cells p1:p3 (p2 - p1) positions
         ;; then repeat from (p1 + (p3 - p2))
         ;;(format t "~&SHIFT=~D P1=~D P2=~D P3=~D~%" shift p1 p2 p3)
         (shift-block-swap array shift p1 p3)
         (when (not p3)
           ;; nothing more to do
           (return))
         (incf p1 (- p3 p2)))))))


(defun remove-column (cells n)
  (shift-block-swap cells 1 n))
|#
