;;; -*- mode: LISP; package: CLONES.CLIM -*-

(in-package #:clones.clim)



(defvar *default-theme*
  (make-instance 'rectangle-theme))



;;; Application frame

(define-application-frame clones (standard-application-frame field)
  ((theme :initarg :theme
          :initform *default-theme*)
   (hover-cells :initform nil)
   (record)
   (score :initarg :score
          :accessor score
          :type integer
          :initform 0)
   (high-scores :initarg :high-scores
                :accessor high-scores
                :initform (read-scores)))
  (:menu-bar menu-command-table)
  (:panes
   (area :application
         :display-function 'draw-field
         :incremental-redisplay t
         :display-after-commands nil
         :display-time t
         :background *background-ink*
         :scroll-bars nil
         ;; FIXME Width and height should be set programmatically
         ;; depending on theme and field dimensions.
         :width :compute ;; (* *cell-width* *field-width*)
         :height :compute ;;(* *cell-height* *field-height*)
         )
   (selected-display text-field
                     :value "Selected:"
                     :editable-p nil)
   (score-display text-field
                  :value "Score:"
                  :editable-p nil))
  (:layouts
   (default
     (vertically ()
       (horizontally ()
                     (1/2 selected-display)
                     (1/2 score-display))
       area))))


(defmethod default-frame-top-level :before ((frame clones) &rest options)
  (declare (ignore options))
  (initialize-theme (slot-value frame 'theme) (get-frame-pane frame 'area)))



;;; Commands


(make-command-table 'menu-game-command-table
                    :errorp nil
                    :menu '(("New" :command com-new)
                            ("Scores" :command com-scores)
                            ("Quit" :command com-quit)))


(make-command-table 'menu-command-table
                    :errorp nil
                    :menu '(("Game" :menu menu-game-command-table)))


(define-clones-command (com-new)
  ()
  (with-slots (score hover-cells record) *application-frame*
    (clones::initialize-field *application-frame*)
    (setf hover-cells nil
          (score *application-frame*) 0)
    (let ((stream (find-pane-named *application-frame* 'area)))
      (clear-output-record record)
      (draw-field *application-frame* stream t))))


(defun show-scores (frame)
  ;; FIXME: Should do this some other way, because this suspends the
  ;; calling thread.
  (run-frame-top-level (make-application-frame 'clones-scores :scores (high-scores frame))
                       :calling-frame frame))


(define-clones-command (com-scores)
  ()
  (show-scores *application-frame*))


(define-clones-command (com-quit)
  ()
  (frame-exit *application-frame*))



;;; Drawing


(eval-when (:compile-toplevel :load-toplevel :execute)
  (define-presentation-type figure ()))


(defun draw-figure (frame i j stream)
  (with-slots (theme) frame
    (let ((cell (cell-at frame i j)))
      (updating-output (stream :cache-value (cell-color cell)
                               :unique-id cell)
        (with-output-as-presentation (stream cell 'figure :single-box nil)
          (draw-cell theme cell stream))))))


(defun draw-field (field &optional (stream *standard-output*) redraw)
  (let ((record (updating-output (stream :all-new redraw)
                  (dotimes (i (field-width field))
                    (dotimes (j (field-height field))
                      (draw-figure field i j stream))))))
    (setf (slot-value field 'record) record)
    record))


(defun draw-cells (list field stream)
  (updating-output (stream)
    (dolist (cell list)
      (draw-figure field (cell-x cell) (cell-y cell) stream))))


(define-presentation-method highlight-presentation ((type figure) record stream state)
  (when (eq :highlight state)
    (let* ((cell (presentation-object record))
           (frame (cell-field cell)))
      (with-slots (hover-cells record) frame
        (cond ((cell-empty-p cell)
               (show-selected-count frame 0)
               (set-region-highlight hover-cells nil)
               (draw-cells hover-cells frame stream)
               (setf hover-cells nil))

              ((not (cell-in-list-p cell hover-cells))
               (set-region-highlight hover-cells nil)
               (draw-cells hover-cells frame stream)
          
               (setf hover-cells (region-at frame (cell-x cell) (cell-y cell)))
               (set-region-highlight hover-cells t)
               (draw-cells hover-cells frame stream)
               (show-selected-count frame (length hover-cells))))))))



;;; Field interaction


(define-presentation-to-command-translator select-figure
  (figure com-select-figure clones
          :gesture :select
          :echo nil)
  (object)
  (list object))


(defun check-game-end (frame)
  (unless (moves-left-p frame)
    (when (field-empty-p frame)
      (incf (score frame) 1000))
    (push (make-instance 'score
                         :amount (score frame)
                         ;; TODO: Ask user's name
                         :name "CLIM user")
          (high-scores frame))
    (setf (high-scores frame)
          (write-scores (high-scores frame)))
    ;; TODO: Shoud show somewhere that the game has ended!
    ;; For now, pop up the scores window.
    (show-scores frame)))


(define-clones-command (com-select-figure :name nil) ((cell 'cell))
  (let ((frame (cell-field cell)))
    (with-slots (hover-cells theme record) frame
      (let ((list (if (cell-empty-p cell)
                      nil
                      (region-at frame (cell-x cell) (cell-y cell)))))
        (when list
          (setf hover-cells nil)
          (remove-cells frame list)
          (normalize-field frame)
          (let ((stream (find-pane-named frame 'area)))
            (clear-output-record record)
            (draw-field frame stream t))
          (incf (score frame) (calculate-score list))
          (check-game-end frame))))))


(defun calculate-score (list)
  (expt (- (length list) 2) 2))


(defmethod (setf score) :after (value (frame clones))
  (setf (gadget-value (find-pane-named frame 'score-display))
        (format nil "Score: ~6D" value)))


(defun show-selected-count (frame count)
  (setf (gadget-value (find-pane-named frame 'selected-display))
        (format nil "Selected: ~4D" count)))



;;; Other stuff


(defun pick-random-theme ()
  (let ((themes (load-time-value
                 `#(,(make-instance 'rectangle-theme)
                    ,(make-instance 'circle-theme)
                    ,(make-instance 'pixmap-theme :name "simple")
                    ,(make-instance 'pixmap-theme :name "square")))))
    (svref themes (random (length themes)))))


(defun clones (&optional (theme (pick-random-theme)))
  (run-frame-top-level (make-application-frame 'clones :theme theme)))
